#include <stdio.h>
#include "iodefine.h"


/* LCD for the RSD is:

   14 D7  DLCDD7  P9.7  pass MSN first, then LSN.
   13 D6  DLCDD6  P9.6
   12 D5  DLCDD5  P9.5
   11 D4  DLCDD4  P9.4
   10 -
    9 -
    8 -
    7 -
    6 En  DLCDE   P8.5  Pulse H
    5 RW  Vss
    4 RS  DLCDRS  P8.4
    3 Vo  -
    2 Vdd Vdd
    1 Vss Vss

    E RS Func
    0 -  disabled
    1 0  instruction write
    1 1  data write

     0  1  2  3  4  5  6  7  addresses
    40 41 42 43 44 45 46 47
*/

#define DLCDD	PORT9.DR.BYTE
#define DLCDE	PORT8.DR.BIT.B5
#define DLCDRS	PORT8.DR.BIT.B4

static inline void
wait_ns_1 (int n)
{
  /* This takes 1 cycle for the add, and 3 cycles for the branch.  At
     100 MHz, that's 40 nS per iteration.  */
  __asm__ __volatile__ ("\n\
1:	add	#-1, %0\n\
	bgt	1b"
	 : "=r" (n) : "0" (n));
}

/* Scale each of these to the number of iterations of the above loop,
   rounding up - delays are minimums, not maximums.  */
#define wait_ns(x) wait_ns_1 ((x + 39) / 40)
#define wait_us(x) wait_ns_1 (((x*1000) + 39) / 40)
#define wait_ms(x) wait_ns_1 (((x*1000000) + 39) / 40)

static inline void
lcd_nibble (int n)
{
  DLCDD = n;
  DLCDE = 1;
  wait_ns (450);
  DLCDE = 0;
  wait_ns (450);
}

static void
lcd_command (int c)
{
  DLCDRS = 0;
  lcd_nibble (c);
  lcd_nibble (c << 4);
}

static void
lcd_data (int d)
{
  DLCDRS = 1;
  lcd_nibble (d);
  lcd_nibble (d << 4);
  wait_us (40);
}

void
lcd_init (void)
{
  PORT8.DDR.BIT.B5 = 1;
  PORT8.DDR.BIT.B4 = 1;
  PORT9.DDR.BYTE |= 0xf0;

  DLCDE = 0;
  DLCDRS = 0;
  wait_ms (31);

  DLCDRS = 0;

  /* We don't know the state of the LCD.  There are three states we
     could be in:

     1. 8-bit mode.  We issue three "8-bit" commands, then a 4-bit
       command.

     2. 4-bit mode, in sync.  We issue two 8-bit commands, then a
        4-bit command.

     3. 4-bit mode, between nibbles.  We issue garbage followed by an
        8-bit command, followed by a 4-bit command.
  */
  /*                    case 1	case 2	case 3 */
  lcd_nibble (0x30); /* 8:8bit	4:8bitH	4:garbage */
  wait_us (40);
  lcd_nibble (0x30); /* 8:8bit	4:8bitL	4:8bitH */
  wait_us (40);
  lcd_nibble (0x30); /* 8:8bit	8:8bit	4:8bitL */
  wait_us (40);
  /* We now know what state the LCD is in - we're in 8-bit mode.  */
  lcd_nibble (0x20); /* Switch to 4-bit interface, from 8-bit mode.  */
  wait_us (40);
  /* We now know we're in 4-bit mode.  */
  lcd_command (0x2c); /* LCD on, two rows */
  wait_us (40);

  lcd_command (0x0e); /* display on, cursor on, blink off */
  wait_ms (4);

  lcd_command (0x06); /* increment, shift off */
  wait_us (40);
  lcd_command (0x01); /* clear screen */
  wait_ms (4);
}

void
lcd_putstr (int row, int col, char *str)
{
  lcd_command (0x80 + row * 0x40 + col);
  wait_us (39);
  while (*str)
    lcd_data (*str++);
}
