#include <stdio.h>
#include <stdint.h>
#include "iodefine.h"
#include "timer.h"
#include "sdram.h"

#define SDRAM ((void *)0x08000000)
#define SDRAM_MEGS 8

static int err_count = 0;

static int old_sec = -1;
static const char *old_test = NULL;
#define STATUS(test, ptr) if (timer_ms/1000 != old_sec || test != old_test) status(test, ptr)

static void
status (const char *test, volatile uint16_t *ptr)
{
  old_sec = timer_ms / 1000;
  old_test = test;
  printf("  %08x %s\033[K\n", ptr, test);
}

static void
err (volatile uint16_t *m, uint16_t w, uint16_t s)
{
  int i;
  printf("%08x : want %04x saw %04x  : ", m, w, s);
  for (i=0; i<8; i++)
    {
      volatile uint16_t *mp = m + i - 4;
      if (i == 4)
	printf("\033[36m");
      printf(" %02x:%04x", (int)mp & 0xff, *mp);
      if (i == 4)
	printf("\033[0m");
    }
  printf("\n");
  if (++err_count == 100)
    {
      printf("Too many errors, stopped.\n");
      fflush (stdout);
      while (1)
	;
    }
}

/*--------------------------------------------------------------------------*/

static unsigned long long rand_next = 1;

static unsigned long
sd_rand(void)
{
  /* This multiplier was obtained from Knuth, D.E., "The Art of
     Computer Programming," Vol 2, Seminumerical Algorithms, Third
     Edition, Addison-Wesley, 1998, p. 106 (line 26) & p. 108 */
  rand_next = rand_next * 6364136223846793005LL + 1;
  /* was: next = next * 0x5deece66dLL + 11; */
  return (unsigned long)(rand_next >> 32);
}

static void
sd_srand(unsigned seed)
{
  rand_next = seed;
}

/*--------------------------------------------------------------------------*/

static void
mt_fill (volatile uint16_t *mem, int len, uint16_t pattern)
{
  char buf[30];
  int i;

  sprintf (buf, "fill[%04x]/wr", pattern);
  for (i=0; i<len; i++)
    {
      STATUS (buf, mem+i);
      mem[i] = pattern;
    }

  sprintf (buf+1, "fill[%04x]/rd", pattern);
  for (i=0; i<len; i++)
    {
      uint16_t m = mem[i];
      STATUS (buf+1, mem+i);
      if (m != pattern)
	err (mem+i, pattern, m);
    }
}

static void
mt_rand (volatile uint16_t *mem, int len)
{
  int i;
  uint16_t p;
  unsigned long long seed = rand_next;

  for (i=0; i<len; i++)
    {
      STATUS ("rand/wr", mem+i);
      p = sd_rand ();
      mem[i] = p;
    }

  rand_next = seed;
  for (i=0; i<len; i++)
    {
      uint16_t m = mem[i];
      STATUS ("rand/rd", mem+i);
      p = sd_rand ();
      if (m != p)
	err (mem+i, p, m);
    }
}

static void
mt_addr (volatile uint16_t *mem, int len)
{
  int i, s;
  char buf[30];

  for (s=0; s<9; s++)
    {
      sprintf (buf, "addr>>%d/wr", s);
      for (i=0; i<len; i++)
	{
	  STATUS (buf, mem+i);
	  mem[i] = (i>>s) & 0xffff;
	}
      sprintf (buf+1, "addr>>%d/rd", s);
      for (i=0; i<len; i++)
	{
	  uint16_t p = (i>>s) & 0xffff;
	  uint16_t m = mem[i];

	  STATUS (buf+1, mem+i);

	  if (m != p)
	    err (mem+i, p, m);
	}
    }
}

static void
mt_timing (volatile uint16_t *mem, int len)
{
  int i, s;

  printf("WR timing: ");
  s = timer_ms;
  sstr_w (mem, 0, len);
  i = timer_ms;
  printf ("%d writes in %d msec (%d kt/s)\n", len, i - s, len / (i-s));

  /* We know it's full of zeros :-)  */
  printf("RD timing: ");
  s = timer_ms;
  suntil_w (mem, 0xff, len);
  i = timer_ms;
  printf ("%d reads in %d msec (%d kt/s)\n", len, i - s, len / (i-s));
}

static void
memtest (int pass, volatile uint16_t *mem, int len)
{
  int i;
  printf("Pass %d testing %08x to %08x\n",
	 pass,
	 mem,
	 mem + len - sizeof (*mem));

  mt_timing (mem, len);
  mt_rand (mem, len);
  mt_addr (mem, len);

  mt_fill (mem, len, 0x0000);
  mt_fill (mem, len, 0xffff);
  mt_fill (mem, len, 0xaaaa);
  mt_fill (mem, len, 0x5555);

  for (i=0; i<16; i++)
    mt_fill (mem, len, 1 << i);
  for (i=0; i<16; i++)
    mt_fill (mem, len, ~(1 << i));
}

static unsigned long
size_sdram (void *sdram_v)
{
  unsigned char *sdram = (unsigned char *)sdram_v;
  unsigned long *sdram_l;
  unsigned long i, b;

  for (i = 128 * 1024 * 1024; i > 4; i >>= 1)
    {
      sdram_l =  (unsigned long *) (sdram + i);
      *sdram_l = i;
    }
  sdram_l =  (unsigned long *) sdram;
  return *sdram_l;
}

void
main ()
{
  int i = 0;
  unsigned long bytes;

  SYSTEM.SCKCR.LONG = 0x00010100;

  setbuf (stdout, NULL);

  timer_init ();
  tty_init ();
  tty_puts("\nOSBoard SDRAM tests\n");

  sdram_init ();

  bytes = sdram_size ();
  printf("sdram size = %lu (%d MB)\n", bytes, bytes/1024/1024);

  i = 0;
  while (1)
    {
      i ++;
      memtest (i,
	       (volatile uint16_t *)sdram_base (),
	       sdram_size () / sizeof(uint16_t));
    }
}
